#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['brutalmaze']

package_data = \
{'': ['*'], 'brutalmaze': ['soundfx/*']}

install_requires = \
['appdirs', 'palace', 'pygame>=1.9', 'setuptools']

entry_points = \
{'console_scripts': ['brutalmaze = brutalmaze.game:main']}

setup(name='brutalmaze',
      version='0.9.2',
      description="Minimalist thrilling shoot 'em up game with minimalist art style",
      author='Nguyễn Gia Phong',
      author_email='mcsinyx@disroot.org',
      url='https://github.com/McSinyx/brutalmaze',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
